# Copyright (C) 2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

from kafka import KafkaConsumer

from .serializers import kafka_to_key, kafka_to_value
from swh.journal import DEFAULT_PREFIX

logger = logging.getLogger(__name__)


# Only accepted offset reset policy accepted
ACCEPTED_OFFSET_RESET = ['earliest', 'latest']

# Only accepted object types
ACCEPTED_OBJECT_TYPES = [
    'content',
    'directory',
    'revision',
    'release',
    'snapshot',
    'origin',
    'origin_visit'
]


class JournalClient:
    """A base client for the Software Heritage journal.

    The current implementation of the journal uses Apache Kafka
    brokers to publish messages under a given topic prefix, with each
    object type using a specific topic under that prefix. If the 'prefix'
    argument is None (default value), it will take the default value
    'swh.journal.objects'.

    Clients subscribe to events specific to each object type as listed in the
    `object_types` argument (if unset, defaults to all accepted objet types).

    Clients can be sharded by setting the `group_id` to a common
    value across instances. The journal will share the message
    throughput across the nodes sharing the same group_id.

    Messages are processed by the `process_objects` method in batches
    of maximum `max_messages`.

    Any other named argument is passed directly to KafkaConsumer().

    """
    def __init__(
            self, brokers, group_id, prefix=None, object_types=None,
            max_messages=0, auto_offset_reset='earliest', **kwargs):
        if prefix is None:
            prefix = DEFAULT_PREFIX
        if object_types is None:
            object_types = ACCEPTED_OBJECT_TYPES
        if auto_offset_reset not in ACCEPTED_OFFSET_RESET:
            raise ValueError(
                'Option \'auto_offset_reset\' only accept %s.' %
                ACCEPTED_OFFSET_RESET)

        for object_type in object_types:
            if object_type not in ACCEPTED_OBJECT_TYPES:
                raise ValueError(
                    'Option \'object_types\' only accepts %s.' %
                    ACCEPTED_OFFSET_RESET)

        self.consumer = KafkaConsumer(
            bootstrap_servers=brokers,
            key_deserializer=kafka_to_key,
            value_deserializer=kafka_to_value,
            auto_offset_reset=auto_offset_reset,
            enable_auto_commit=False,
            group_id=group_id,
            **kwargs)

        self.consumer.subscribe(
            topics=['%s.%s' % (prefix, object_type)
                    for object_type in object_types],
        )

        self.max_messages = max_messages
        self._object_types = object_types

    def process(self, worker_fn):
        """Polls Kafka for a batch of messages, and calls the worker_fn
        with these messages.

        Args:
            worker_fn Callable[Dict[str, List[dict]]]: Function called with
                                                       the messages as
                                                       argument.
        """
        nb_messages = 0
        polled = self.consumer.poll()
        for (partition, messages) in polled.items():
            object_type = partition.topic.split('.')[-1]
            # Got a message from a topic we did not subscribe to.
            assert object_type in self._object_types, object_type

            worker_fn({object_type: [msg.value for msg in messages]})

            nb_messages += len(messages)

        self.consumer.commit()
        return nb_messages



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import functools
import random
from subprocess import Popen
from typing import Tuple

import dateutil
from kafka import KafkaProducer

from swh.storage import get_storage
from swh.storage.in_memory import ENABLE_ORIGIN_IDS

from swh.journal.client import JournalClient
from swh.journal.serializers import key_to_kafka, value_to_kafka
from swh.journal.replay import process_replay_objects

from .conftest import OBJECT_TYPE_KEYS
from .utils import MockedJournalClient, MockedKafkaWriter


def test_storage_play(
        kafka_prefix: str,
        kafka_server: Tuple[Popen, int]):
    (_, port) = kafka_server
    kafka_prefix += '.swh.journal.objects'

    storage = get_storage('memory', {})

    producer = KafkaProducer(
        bootstrap_servers='localhost:{}'.format(port),
        key_serializer=key_to_kafka,
        value_serializer=value_to_kafka,
        client_id='test producer',
    )

    now = datetime.datetime.now(tz=datetime.timezone.utc)

    # Fill Kafka
    nb_sent = 0
    nb_visits = 0
    for (object_type, (_, objects)) in OBJECT_TYPE_KEYS.items():
        topic = kafka_prefix + '.' + object_type
        for object_ in objects:
            key = bytes(random.randint(0, 255) for _ in range(40))
            object_ = object_.copy()
            if object_type == 'content':
                object_['ctime'] = now
            elif object_type == 'origin_visit':
                nb_visits += 1
                object_['visit'] = nb_visits
            producer.send(topic, key=key, value=object_)
            nb_sent += 1

    # Fill the storage from Kafka
    config = {
        'brokers': 'localhost:%d' % kafka_server[1],
        'group_id': 'replayer',
        'prefix': kafka_prefix,
        'max_messages': nb_sent,
    }
    replayer = JournalClient(**config)
    worker_fn = functools.partial(process_replay_objects, storage=storage)
    nb_inserted = 0
    while nb_inserted < nb_sent:
        nb_inserted += replayer.process(worker_fn)
    assert nb_sent == nb_inserted

    # Check the objects were actually inserted in the storage
    assert OBJECT_TYPE_KEYS['revision'][1] == \
        list(storage.revision_get(
            [rev['id'] for rev in OBJECT_TYPE_KEYS['revision'][1]]))
    assert OBJECT_TYPE_KEYS['release'][1] == \
        list(storage.release_get(
            [rel['id'] for rel in OBJECT_TYPE_KEYS['release'][1]]))

    origins = list(storage.origin_get(
            [orig for orig in OBJECT_TYPE_KEYS['origin'][1]]))
    assert OBJECT_TYPE_KEYS['origin'][1] == \
        [{'url': orig['url'], 'type': orig['type']} for orig in origins]
    for origin in origins:
        expected_visits = [
            {
                **visit,
                'origin': origin['id'],
                'date': dateutil.parser.parse(visit['date']),
            }
            for visit in OBJECT_TYPE_KEYS['origin_visit'][1]
            if visit['origin']['url'] == origin['url']
            and visit['origin']['type'] == origin['type']
        ]
        actual_visits = list(storage.origin_visit_get(origin['id']))
        for visit in actual_visits:
            del visit['visit']  # opaque identifier
        assert expected_visits == actual_visits

    contents = list(storage.content_get_metadata(
            [cont['sha1'] for cont in OBJECT_TYPE_KEYS['content'][1]]))
    assert None not in contents
    assert contents == OBJECT_TYPE_KEYS['content'][1]


def test_write_replay_legacy_origin_visit1():
    """Test origin_visit when the 'origin' is just a string."""
    queue = []
    replayer = MockedJournalClient(queue)
    writer = MockedKafkaWriter(queue)

    # Note that flipping the order of these two insertions will crash
    # the test, because the legacy origin_format does not allow to create
    # the origin when needed (type is missing)
    now = datetime.datetime.now()
    writer.send('origin', 'foo', {
        'url': 'http://example.com/',
        'type': 'git',
    })
    writer.send('origin_visit', 'foo', {
        'visit': 1,
        'origin': 'http://example.com/',
        'date': now,
    })

    queue_size = sum(len(partition)
                     for batch in queue
                     for partition in batch.values())

    storage = get_storage('memory', {})
    worker_fn = functools.partial(process_replay_objects, storage=storage)
    nb_messages = 0
    while nb_messages < queue_size:
        nb_messages += replayer.process(worker_fn)

    visits = list(storage.origin_visit_get('http://example.com/'))

    if ENABLE_ORIGIN_IDS:
        assert visits == [{
                'visit': 1,
                'origin': 1,
                'date': now,
            }]
    else:
        assert visits == [{
            'visit': 1,
            'origin': {'url': 'http://example.com/'},
            'date': now,
        }]


def test_write_replay_legacy_origin_visit2():
    """Test origin_visit when 'type' is missing."""
    queue = []
    replayer = MockedJournalClient(queue)
    writer = MockedKafkaWriter(queue)

    now = datetime.datetime.now()
    writer.send('origin', 'foo', {
        'url': 'http://example.com/',
        'type': 'git',
    })
    writer.send('origin_visit', 'foo', {
        'visit': 1,
        'origin': {
            'url': 'http://example.com/',
            'type': 'git',
        },
        'date': now,
    })

    queue_size = sum(len(partition)
                     for batch in queue
                     for partition in batch.values())

    storage = get_storage('memory', {})
    worker_fn = functools.partial(process_replay_objects, storage=storage)
    nb_messages = 0
    while nb_messages < queue_size:
        nb_messages += replayer.process(worker_fn)

    visits = list(storage.origin_visit_get('http://example.com/'))

    if ENABLE_ORIGIN_IDS:
        assert visits == [{
            'visit': 1,
            'origin': 1,
            'date': now,
            'type': 'git',
        }]
    else:
        assert visits == [{
            'visit': 1,
            'origin': {'url': 'http://example.com/'},
            'date': now,
            'type': 'git',
        }]



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

from kafka import KafkaProducer

from swh.model.hashutil import DEFAULT_ALGORITHMS

from .serializers import key_to_kafka, value_to_kafka

logger = logging.getLogger(__name__)


class DirectKafkaWriter:
    """This class is instantiated and used by swh-storage to write incoming
    new objects to Kafka before adding them to the storage backend
    (eg. postgresql) itself."""
    def __init__(self, brokers, prefix, client_id):
        self._prefix = prefix

        self.producer = KafkaProducer(
            bootstrap_servers=brokers,
            key_serializer=key_to_kafka,
            value_serializer=value_to_kafka,
            client_id=client_id,
        )

    def send(self, topic, key, value):
        self.producer.send(topic=topic, key=key, value=value)

    def _get_key(self, object_type, object_):
        if object_type in ('revision', 'release', 'directory', 'snapshot'):
            return object_['id']
        elif object_type == 'content':
            return object_['sha1']  # TODO: use a dict of hashes
        elif object_type == 'skipped_content':
            return {
                hash: object_[hash]
                for hash in DEFAULT_ALGORITHMS
            }
        elif object_type == 'origin':
            return {'url': object_['url'], 'type': object_['type']}
        elif object_type == 'origin_visit':
            return {
                'origin': object_['origin'],
                'date': str(object_['date']),
            }
        else:
            raise ValueError('Unknown object type: %s.' % object_type)

    def _sanitize_object(self, object_type, object_):
        if object_type == 'origin_visit':
            return {
                **object_,
                'date': str(object_['date']),
            }
        elif object_type == 'origin':
            assert 'id' not in object_
        return object_

    def write_addition(self, object_type, object_):
        topic = '%s.%s' % (self._prefix, object_type)
        key = self._get_key(object_type, object_)
        object_ = self._sanitize_object(object_type, object_)
        logger.debug('topic: %s, key: %s, value: %s', topic, key, object_)
        self.send(topic, key=key, value=object_)

    write_update = write_addition

    def write_additions(self, object_type, objects):
        for object_ in objects:
            self.write_addition(object_type, object_)



# Copyright (C) 2016-2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import click
import functools
import logging
import os

from swh.core import config
from swh.core.cli import CONTEXT_SETTINGS
from swh.storage import get_storage
from swh.objstorage import get_objstorage

from swh.journal.client import JournalClient
from swh.journal.replay import process_replay_objects
from swh.journal.replay import process_replay_objects_content
from swh.journal.backfill import JournalBackfiller


@click.group(name='journal', context_settings=CONTEXT_SETTINGS)
@click.option('--config-file', '-C', default=None,
              type=click.Path(exists=True, dir_okay=False,),
              help="Configuration file.")
@click.pass_context
def cli(ctx, config_file):
    """Software Heritage Journal tools.

    The journal is a persistent logger of changes to the archive, with
    publish-subscribe support.

    """
    if not config_file:
        config_file = os.environ.get('SWH_CONFIG_FILENAME')

    if config_file:
        if not os.path.exists(config_file):
            raise ValueError('%s does not exist' % config_file)
        conf = config.read(config_file)
    else:
        conf = {}

    ctx.ensure_object(dict)

    log_level = ctx.obj.get('log_level', logging.INFO)
    logging.root.setLevel(log_level)
    logging.getLogger('kafka').setLevel(logging.INFO)

    ctx.obj['config'] = conf


def get_journal_client(ctx, **kwargs):
    conf = ctx.obj['config'].get('journal', {})
    conf.update({k: v for (k, v) in kwargs.items() if v not in (None, ())})
    if not conf.get('brokers'):
        ctx.fail('You must specify at least one kafka broker.')
    if not isinstance(conf['brokers'], (list, tuple)):
        conf['brokers'] = [conf['brokers']]
    return JournalClient(**conf)


@cli.command()
@click.option('--max-messages', '-m', default=None, type=int,
              help='Maximum number of objects to replay. Default is to '
                   'run forever.')
@click.option('--broker', 'brokers', type=str, multiple=True,
              help='Kafka broker to connect to. '
                   '(deprecated, use the config file instead)')
@click.option('--prefix', type=str, default=None,
              help='Prefix of Kafka topic names to read from. '
                   '(deprecated, use the config file instead)')
@click.option('--group-id', type=str,
              help='Name of the group id for reading from Kafka. '
                   '(deprecated, use the config file instead)')
@click.pass_context
def replay(ctx, brokers, prefix, group_id, max_messages):
    """Fill a Storage by reading a Journal.

    There can be several 'replayers' filling a Storage as long as they use
    the same `group-id`.
    """
    logger = logging.getLogger(__name__)
    conf = ctx.obj['config']
    try:
        storage = get_storage(**conf.pop('storage'))
    except KeyError:
        ctx.fail('You must have a storage configured in your config file.')

    client = get_journal_client(
        ctx, brokers=brokers, prefix=prefix, group_id=group_id)
    worker_fn = functools.partial(process_replay_objects, storage=storage)

    try:
        nb_messages = 0
        while not max_messages or nb_messages < max_messages:
            nb_messages += client.process(worker_fn)
            logger.info('Processed %d messages.' % nb_messages)
    except KeyboardInterrupt:
        ctx.exit(0)
    else:
        print('Done.')


@cli.command()
@click.argument('object_type')
@click.option('--start-object', default=None)
@click.option('--end-object', default=None)
@click.option('--dry-run', is_flag=True, default=False)
@click.pass_context
def backfiller(ctx, object_type, start_object, end_object, dry_run):
    """Run the backfiller

    The backfiller list objects from a Storage and produce journal entries from
    there.

    Typically used to rebuild a journal or compensate for missing objects in a
    journal (eg. due to a downtime of this later).

    The configuration file requires the following entries:
    - brokers: a list of kafka endpoints (the journal) in which entries will be
               added.
    - storage_dbconn: URL to connect to the storage DB.
    - prefix: the prefix of the topics (topics will be <prefix>.<object_type>).
    - client_id: the kafka client ID.

    """
    conf = ctx.obj['config']
    backfiller = JournalBackfiller(conf)
    try:
        backfiller.run(
            object_type=object_type,
            start_object=start_object, end_object=end_object,
            dry_run=dry_run)
    except KeyboardInterrupt:
        ctx.exit(0)


@cli.command()
@click.option('--concurrency', type=int,
              default=8,
              help='Concurrentcy level.')
@click.option('--broker', 'brokers', type=str, multiple=True,
              help='Kafka broker to connect to.'
                   '(deprecated, use the config file instead)')
@click.option('--prefix', type=str, default=None,
              help='Prefix of Kafka topic names to read from.'
                   '(deprecated, use the config file instead)')
@click.option('--group-id', type=str,
              help='Name of the group id for reading from Kafka.'
                   '(deprecated, use the config file instead)')
@click.pass_context
def content_replay(ctx, concurrency, brokers, prefix, group_id):
    """Fill a destination Object Storage (typically a mirror) by reading a Journal
    and retrieving objects from an existing source ObjStorage.

    There can be several 'replayers' filling a given ObjStorage as long as they
    use the same `group-id`.

    This service retrieves object ids to copy from the 'content' topic. It will
    only copy object's content if the object's description in the kafka
    nmessage has the status:visible set.
    """
    logger = logging.getLogger(__name__)
    conf = ctx.obj['config']
    try:
        objstorage_src = get_objstorage(**conf.pop('objstorage_src'))
    except KeyError:
        ctx.fail('You must have a source objstorage configured in '
                 'your config file.')
    try:
        objstorage_dst = get_objstorage(**conf.pop('objstorage_dst'))
    except KeyError:
        ctx.fail('You must have a destination objstorage configured '
                 'in your config file.')

    client = get_journal_client(
        ctx, brokers=brokers, prefix=prefix, group_id=group_id,
        object_types=('content',))
    worker_fn = functools.partial(process_replay_objects_content,
                                  src=objstorage_src,
                                  dst=objstorage_dst,
                                  concurrency=concurrency)

    try:
        nb_messages = 0
        while True:
            nb_messages += client.process(worker_fn)
            logger.info('Processed %d messages.' % nb_messages)
    except KeyboardInterrupt:
        ctx.exit(0)
    else:
        logger.info('Done.')


def main():
    logging.basicConfig()
    return cli(auto_envvar_prefix='SWH_JOURNAL')


if __name__ == '__main__':
    main()



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from time import time
import logging
from concurrent.futures import ThreadPoolExecutor

from swh.storage import HashCollision
from swh.model.hashutil import hash_to_hex
from swh.objstorage.objstorage import ID_HASH_ALGO
from swh.core.statsd import statsd

logger = logging.getLogger(__name__)


def process_replay_objects(all_objects, *, storage):
    for (object_type, objects) in all_objects.items():
        _insert_objects(object_type, objects, storage)


def _insert_objects(object_type, objects, storage):
    if object_type == 'content':
        # TODO: insert 'content' in batches
        for object_ in objects:
            try:
                storage.content_add_metadata([object_])
            except HashCollision as e:
                logger.error('Hash collision: %s', e.args)
    elif object_type in ('directory', 'revision', 'release',
                         'snapshot', 'origin'):
        # TODO: split batches that are too large for the storage
        # to handle?
        method = getattr(storage, object_type + '_add')
        method(objects)
    elif object_type == 'origin_visit':
        for visit in objects:
            if isinstance(visit['origin'], str):
                # old format; note that it will crash with the pg and
                # in-mem storages if the origin is not already known,
                # but there is no other choice because we can't add an
                # origin without knowing its type. Non-pg storages
                # don't use a numeric FK internally,
                visit['origin'] = {'url': visit['origin']}
            else:
                storage.origin_add_one(visit['origin'])
                if 'type' not in visit:
                    # old format
                    visit['type'] = visit['origin']['type']

        storage.origin_visit_upsert(objects)
    else:
        logger.warning('Received a series of %s, this should not happen',
                       object_type)


def copy_object(obj_id, src, dst):
    statsd_name = 'swh_journal_content_replayer_%s_duration_seconds'
    try:
        with statsd.timed(statsd_name % 'get'):
            obj = src.get(obj_id)
        with statsd.timed(statsd_name % 'put'):
            dst.add(obj, obj_id=obj_id, check_presence=False)
            logger.debug('copied %s', hash_to_hex(obj_id))
        statsd.increment(
            'swh_journal_content_replayer_bytes_total',
            len(obj))
    except Exception:
        obj = ''
        logger.exception('Failed to copy %s', hash_to_hex(obj_id))
    return len(obj)


def process_replay_objects_content(all_objects, *, src, dst, concurrency=8):
    vol = []
    t0 = time()
    with ThreadPoolExecutor(max_workers=concurrency) as executor:
        for (object_type, objects) in all_objects.items():
            if object_type != 'content':
                logger.warning(
                    'Received a series of %s, this should not happen',
                    object_type)
                continue
            for obj in objects:
                obj_id = obj[ID_HASH_ALGO]
                if obj['status'] == 'visible':
                    fut = executor.submit(copy_object, obj_id, src, dst)
                    fut.add_done_callback(lambda fn: vol.append(fn.result()))
                else:
                    logger.debug('skipped %s (%s)',
                                 hash_to_hex(obj_id), obj['status'])
    dt = time() - t0
    logger.info(
        'processed %s content objects in %.1fsec '
        '(%.1f obj/sec, %.1fMB/sec) - %s failures',
        len(vol), dt,
        len(vol)/dt,
        sum(vol)/1024/1024/dt, len([x for x in vol if not x]))



from collections import namedtuple

from swh.journal.client import JournalClient, ACCEPTED_OBJECT_TYPES
from swh.journal.direct_writer import DirectKafkaWriter
from swh.journal.serializers import (
    key_to_kafka, kafka_to_key, value_to_kafka, kafka_to_value)

FakeKafkaMessage = namedtuple('FakeKafkaMessage', 'key value')
FakeKafkaPartition = namedtuple('FakeKafkaPartition', 'topic')


class MockedKafkaWriter(DirectKafkaWriter):
    def __init__(self, queue):
        self._prefix = 'prefix'
        self.queue = queue

    def send(self, topic, key, value):
        key = kafka_to_key(key_to_kafka(key))
        value = kafka_to_value(value_to_kafka(value))
        partition = FakeKafkaPartition(topic)
        msg = FakeKafkaMessage(key=key, value=value)
        if self.queue and {partition} == set(self.queue[-1]):
            # The last message is of the same object type, groupping them
            self.queue[-1][partition].append(msg)
        else:
            self.queue.append({partition: [msg]})


class MockedKafkaConsumer:
    def __init__(self, queue):
        self.queue = queue
        self.committed = False

    def poll(self):
        return self.queue.pop(0)

    def commit(self):
        if self.queue == []:
            self.committed = True


class MockedJournalClient(JournalClient):
    def __init__(self, queue, object_types=ACCEPTED_OBJECT_TYPES):
        self._object_types = object_types
        self.consumer = MockedKafkaConsumer(queue)



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import functools

from hypothesis import given, settings, HealthCheck
from hypothesis.strategies import lists

from swh.model.hypothesis_strategies import object_dicts
from swh.storage.in_memory import Storage
from swh.storage import HashCollision

from swh.journal.replay import process_replay_objects
from swh.journal.replay import process_replay_objects_content

from .utils import MockedJournalClient, MockedKafkaWriter


@given(lists(object_dicts(), min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_write_replay_same_order_batches(objects):
    queue = []
    replayer = MockedJournalClient(queue)

    storage1 = Storage()
    storage1.journal_writer = MockedKafkaWriter(queue)

    for (obj_type, obj) in objects:
        obj = obj.copy()
        if obj_type == 'origin_visit':
            origin_id = storage1.origin_add_one(obj.pop('origin'))
            if 'visit' in obj:
                del obj['visit']
            storage1.origin_visit_add(origin_id, **obj)
        else:
            method = getattr(storage1, obj_type + '_add')
            try:
                method([obj])
            except HashCollision:
                pass

    queue_size = sum(len(partition)
                     for batch in queue
                     for partition in batch.values())

    storage2 = Storage()
    worker_fn = functools.partial(process_replay_objects, storage=storage2)
    nb_messages = 0
    while nb_messages < queue_size:
        nb_messages += replayer.process(worker_fn)

    assert replayer.consumer.committed

    for attr_name in ('_contents', '_directories', '_revisions', '_releases',
                      '_snapshots', '_origin_visits', '_origins'):
        assert getattr(storage1, attr_name) == getattr(storage2, attr_name), \
            attr_name


# TODO: add test for hash collision


@given(lists(object_dicts(), min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_write_replay_content(objects):
    queue = []
    replayer = MockedJournalClient(queue)

    storage1 = Storage()
    storage1.journal_writer = MockedKafkaWriter(queue)

    for (obj_type, obj) in objects:
        obj = obj.copy()
        if obj_type == 'content':
            storage1.content_add([obj])

    queue_size = sum(len(partition)
                     for batch in queue
                     for partition in batch.values())

    storage2 = Storage()
    worker_fn = functools.partial(process_replay_objects_content,
                                  src=storage1.objstorage,
                                  dst=storage2.objstorage)
    nb_messages = 0
    while nb_messages < queue_size:
        nb_messages += replayer.process(worker_fn)

    assert storage1.objstorage.state == storage2.objstorage.state



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import pytest
import logging
import random
import string

from kafka import KafkaConsumer
from subprocess import Popen
from typing import Tuple, Dict

from pathlib import Path
from pytest_kafka import (
    make_zookeeper_process, make_kafka_server, constants
)

from swh.model.hashutil import hash_to_bytes

from swh.journal.serializers import kafka_to_key, kafka_to_value


CONTENTS = [
    {
        'length': 3,
        'sha1': hash_to_bytes(
            '34973274ccef6ab4dfaaf86599792fa9c3fe4689'),
        'sha1_git': b'foo',
        'blake2s256': b'bar',
        'sha256': b'baz',
        'status': 'visible',
    },
]

COMMITTER = [
    {
        'id': 1,
        'fullname': b'foo',
    },
    {
        'id': 2,
        'fullname': b'bar',
    }
]

REVISIONS = [
    {
        'id': hash_to_bytes('7026b7c1a2af56521e951c01ed20f255fa054238'),
        'message': b'hello',
        'date': {
            'timestamp': {
                'seconds': 1234567891,
                'microseconds': 0,
            },
            'offset': 120,
            'negative_utc': None,
        },
        'committer': COMMITTER[0],
        'author':  COMMITTER[0],
        'committer_date': None,
    },
    {
        'id': hash_to_bytes('368a48fe15b7db2383775f97c6b247011b3f14f4'),
        'message': b'hello again',
        'date': {
            'timestamp': {
                'seconds': 1234567892,
                'microseconds': 0,
            },
            'offset': 120,
            'negative_utc': None,
        },
        'committer': COMMITTER[1],
        'author':  COMMITTER[1],
        'committer_date': None,
    },
]

RELEASES = [
    {
        'id': hash_to_bytes('d81cc0710eb6cf9efd5b920a8453e1e07157b6cd'),
        'name': b'v0.0.1',
        'date': {
            'timestamp': {
                'seconds': 1234567890,
                'microseconds': 0,
            },
            'offset': 120,
            'negative_utc': None,
        },
        'author': COMMITTER[0],
    },
]

ORIGINS = [
    {
        'url': 'https://somewhere.org/den/fox',
        'type': 'git',
    },
    {
        'url': 'https://overtherainbow.org/fox/den',
        'type': 'svn',
    }
]

ORIGIN_VISITS = [
    {
        'origin': ORIGINS[0],
        'date': '2013-05-07 04:20:39.369271+00:00',
        'snapshot': None,  # TODO
        'status': 'ongoing',  # TODO
        'metadata': {'foo': 'bar'},
        'type': 'git',
    },
    {
        'origin': ORIGINS[0],
        'date': '2018-11-27 17:20:39+00:00',
        'snapshot': None,  # TODO
        'status': 'ongoing',  # TODO
        'metadata': {'baz': 'qux'},
        'type': 'git',
    }
]

# From type to tuple (id, <objects instances to test>)
OBJECT_TYPE_KEYS = {
    'content': ('sha1', CONTENTS),
    'revision': ('id', REVISIONS),
    'release': ('id', RELEASES),
    'origin': (None, ORIGINS),
    'origin_visit': (None, ORIGIN_VISITS),
}


KAFKA_ROOT = os.environ.get('SWH_KAFKA_ROOT')
KAFKA_ROOT = KAFKA_ROOT if KAFKA_ROOT else os.path.dirname(__file__) + '/kafka'
if not os.path.exists(KAFKA_ROOT):
    msg = ('Development error: %s must exist and target an '
           'existing kafka installation' % KAFKA_ROOT)
    raise ValueError(msg)

KAFKA_SCRIPTS = Path(KAFKA_ROOT) / 'bin'

KAFKA_BIN = str(KAFKA_SCRIPTS / 'kafka-server-start.sh')
ZOOKEEPER_BIN = str(KAFKA_SCRIPTS / 'zookeeper-server-start.sh')


# Those defines fixtures
zookeeper_proc = make_zookeeper_process(ZOOKEEPER_BIN, scope='session')
os.environ['KAFKA_LOG4J_OPTS'] = \
    '-Dlog4j.configuration=file:%s/log4j.properties' % \
    os.path.dirname(__file__)
kafka_server = make_kafka_server(KAFKA_BIN, 'zookeeper_proc', scope='session')

logger = logging.getLogger('kafka')
logger.setLevel(logging.WARN)


@pytest.fixture(scope='function')
def kafka_prefix():
    return ''.join(random.choice(string.ascii_lowercase) for _ in range(10))


TEST_CONFIG = {
    'consumer_id': 'swh.journal.consumer',
    'object_types': OBJECT_TYPE_KEYS.keys(),
    'max_messages': 1,  # will read 1 message and stops
    'storage': {'cls': 'memory', 'args': {}},
}


@pytest.fixture
def test_config(kafka_server: Tuple[Popen, int],
                kafka_prefix: str):
    """Test configuration needed for producer/consumer

    """
    _, port = kafka_server
    return {
        **TEST_CONFIG,
        'brokers': ['localhost:{}'.format(port)],
        'prefix': kafka_prefix + '.swh.journal.objects',
    }


@pytest.fixture
def consumer(
        kafka_server: Tuple[Popen, int], test_config: Dict) -> KafkaConsumer:
    """Get a connected Kafka consumer.

    """
    kafka_topics = [
        '%s.%s' % (test_config['prefix'], object_type)
        for object_type in test_config['object_types']]
    _, kafka_port = kafka_server
    consumer = KafkaConsumer(
        *kafka_topics,
        bootstrap_servers='localhost:{}'.format(kafka_port),
        consumer_timeout_ms=constants.DEFAULT_CONSUMER_TIMEOUT_MS,
        key_deserializer=kafka_to_key,
        value_deserializer=kafka_to_value,
        auto_offset_reset='earliest',
        enable_auto_commit=True,
        group_id="test-consumer"
    )

    # Enforce auto_offset_reset=earliest even if the consumer was created
    # too soon wrt the server.
    while len(consumer.assignment()) == 0:
        consumer.poll(timeout_ms=20)
    consumer.seek_to_beginning()

    return consumer
